import json

from .node import Node


class Selector(Node):
    def __init__(self, segments=None, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.inherits = True
        self.segments = segments
        self.optional = False
        self.block = None

    def __str__(self):
        try:
            string = "".join(str(self.segments))
            if self.optional:
                string += " !optional"
            return string
        except TypeError:
            return "n/a"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.inherits, self.segments, self.optional

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value == other.value

    def is_placeholder(self):
        if self.value and "$" in self.value[0:2]:
            return True
        else:
            return False

    def clone(self, parent=None, node=None):
        clone = Selector(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        clone.inherits = self.inherits
        clone.value = self.value
        clone.segments = [node.clone(parent, clone) for node in self.segments]
        clone.optional = self.optional
        clone.block = self.block
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Selector",
                "inherits": self.inherits,
                "segments": self.segments,
                "optional": self.optional,
                "val": self.value,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



units = [
    "em",
    "ex",
    "ch",
    "rem",  # relative lengths
    "vw",
    "vh",
    "vmin",
    "vmax",  # relative viewport-percentage lengths
    "cm",
    "mm",
    "in",
    "pt",
    "pc",
    "px",  # absolute lengths
    "deg",
    "grad",
    "rad",
    "turn",  # angles
    "s",
    "ms",  # times
    "Hz",
    "kHz",  # frequencies
    "dpi",
    "dpcm",
    "dppx",
    "x",  # resolutions
    "%",  # percentage type
    "fr",  # grid-layout (http: #www.w3.org/TR/css3-grid-layout/)
]



from stilus.exceptions import StilusError
from stilus.nodes.expression import Expression
from stilus.nodes.unit import Unit
from stilus.utils import assert_type


def range_function(start, stop, step=None, evaluator=None):
    assert_type(start, "unit", "start")
    assert_type(stop, "unit", "stop")
    if step:
        assert_type(step, "unit", "step")
        if step.value == 0:
            raise StilusError(
                'ArgumentError: "step" argument ' "must not be zero"
            )
    else:
        step = Unit(1)
    lst = Expression()
    i = start.value
    while i <= stop.value:
        lst.append(Unit(i, start.type))
        i += step.value
    return lst



from stilus import utils


# todo: the *args is there due to javascript weirdness.
def length(expr, *args, evaluator=None):
    if expr:
        if expr.nodes:
            nodes = utils.unwrap(expr).nodes
            if len(nodes) == 1 and nodes[0].node_name == "objectnode":
                return len(nodes[0])
            else:
                return len(nodes)
        else:
            return 1
    return 0



from stilus.utils import unwrap


def push(expr, *args, evaluator=None):
    expr = unwrap(expr)
    if evaluator:
        for arg in args:
            expr.nodes.append(unwrap(arg).clone())
    return len(expr.nodes)



import json

from .node import Node


class Query(Node):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.nodes = []
        self.type = ""
        self.predicate = ""

    def __str__(self):
        length = len(self.nodes)
        pred = str(self.predicate) + " " if self.predicate else ""
        if self.type:
            type = str(self.type)
            if length > 0:
                type += " and "
        else:
            type = ""
        s = " and ".join([str(e) for e in self.nodes])
        return f"{pred}{type}{s}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.value

    def __eq__(self, other):
        if isinstance(other, Query):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def append(self, feature):
        self.nodes.append(feature)

    def resolved_type(self):
        if self.type:
            if hasattr(self.type, "node_name"):
                return self.type.string
            else:
                return self.type

    def resolved_predicate(self):
        if self.predicate:
            if hasattr(self.predicate, "node_name"):
                return self.predicate.string
            else:
                return self.predicate

    def merge(self, other):
        query = Query()
        p1 = self.resolved_predicate()
        p2 = other.resolved_predicate()
        t1 = self.resolved_type()
        t2 = other.resolved_type()

        if t1:
            t1 = t1
        else:
            t1 = t2
        if t2:
            t2 = t2
        else:
            t2 = t1

        if p1 and p2 and ("not" == p1 ^ "not" == p2):
            if t1 == t2:
                return
            type = t2 if "not" == t1 else t1
            pred = p2 if "not" == p1 else p1
        elif p1 and p2 and ("not" == p1 and "not" == p2):
            if t1 != t2:
                return
            type = t1
            pred = "not"
        elif t1 and t2 and t1 != t2:
            return
        else:
            type = t1
            if p1:
                pred = p1
            else:
                pred = p2

        query.predicate = pred
        query.type = type
        query.nodes.extend(self.nodes)
        query.nodes.extend(other.nodes)
        return query

    def clone(self, parent=None, node=None):
        clone = Query(lineno=self.lineno, column=self.column)
        clone.predicate = self.predicate
        clone.type = self.type
        clone.nodes = [node.clone(parent, clone) for node in self.nodes]
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Query",
                "predicate": self.predicate,
                "type": self.type,
                "nodes": self.nodes,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.selector_parser import SelectorParser
from stilus.utils import assert_string, compile_selectors, unwrap


def parse(selector):
    from stilus.parser import Parser

    parser = Parser(selector, {})
    parser.state.append("selector-parts")
    nodes = parser.stmt_selector()
    for node in nodes:
        node.value = "".join([str(segment) for segment in node.segments])
    return nodes


def push_to_stack(selectors, stack):
    for selector in selectors:
        selector = selector.first()
        assert_string(selector, "selector")
        stack.append(parse(selector.string))


def selector(*args, evaluator=None):
    stack = evaluator.selector_stack()

    if len(args) == 1:
        expr = unwrap(args[0])
        length = len(expr.nodes)

        # selector .a
        if length == 1:
            assert_string(expr.first(), "selector")
            value = expr.first().string
            parsed = SelectorParser(value).parse()["value"]

            if parsed == value:
                return value

            stack.append(parse(value))

        elif length > 1:
            # selector-list = '.a', '.b', '.c'
            # selector(selector-list)
            if expr.is_list:
                push_to_stack(expr.nodes, stack)
            # selector('.a' '.b' '.c')
            else:
                joined = " ".join([node.string for node in expr.nodes])
                stack.append(parse(joined))

    # selector('.a', '.b', '.c')
    elif len(args) > 1:
        push_to_stack(args, stack)

    if stack:
        return ",".join(compile_selectors(stack))
    else:
        return "&"



from stilus.utils import assert_present, assert_type


def operate(op=None, left=None, right=None, evaluator=None):
    assert_type(op, "string", "op")
    assert_present(left, "left")
    assert_present(right, "right")
    return left.operate(op.value, right)



import json
from typing import Type
from stilus.exceptions import StilusError


# todo: move this one to the stilus exceptions
class CoercionError(Exception):
    """CoercionError"""

    pass


current_filename = None  # todo: clean this up


class Node:
    """Node is the root class of all nodes."""

    def __init__(self, value=None, filename=None, lineno=1, column=1):
        self.value = value
        if not filename:
            self.filename = current_filename
        else:
            self.filename = filename
        self.lineno = lineno
        self.column = column
        self.node_name = self.__class__.__name__.lower()
        self.bubbled = False

    def __eq__(self, other):
        if isinstance(other, Node):
            return self.value == other.value
        return False

    def __hash__(self):
        return hash(self.value)

    def hash(self):
        """Return hash."""
        return self.value

    def first(self):
        """Return this node.
        :return: This node.
        """
        return self

    def to_boolean(self):
        """Return True.
        :return: True
        """
        from .boolean import Boolean

        return Boolean(True)

    def to_expression(self):
        """Return expression or wrap this node in an expression
        :return:
        """
        from stilus.nodes.expression import Expression

        if self.node_name == "expression":
            return self
        expression = Expression()
        expression.append(self)
        return expression

    def evaluate(self):
        """Nodes by default evaulate to themselves.
        :return: A Node.
        """
        from visitor.evaluator import Evaluator

        return Evaluator(self).evaluate()

    def one_is_unit(self, other):
        return self.node_name == "unit" or other.node_name == "unit"

    def operate(  # noqa: C901
        self, op: str, right: Type["Node"], value=None
    ) -> "Node":
        """Operate on ``right`` with the given ``op``."""
        from .boolean import Boolean

        if op == "is a":
            if "string" == right.first().node_name:
                return Boolean(self.node_name == right.value)
            else:
                raise Exception(
                    f'"is a" expects a string, ' f"got {right.toString}"
                )
        elif op == "==":
            from stilus import utils

            if utils.is_number(self) and utils.is_number(right):
                return Boolean(utils.get_value(self) == utils.get_value(right))
            return Boolean(self.hash() == right.hash())
        elif op == "!=":
            return Boolean(self.hash() != right.hash())
        elif op == ">=":
            return Boolean(self.hash() >= right.hash())
        elif op == "<=":
            return Boolean(self.hash() <= right.hash())
        elif op == ">":
            if self.one_is_unit(right):
                return Boolean(float(self.hash()) > float(right.hash()))
            return Boolean(self.hash() > right.hash())
        elif op == "<":
            return Boolean(self.hash() < right.hash())
        elif op == "||":
            return self if self.to_boolean().value is True else right
        elif op == "in":
            from stilus import utils

            values = utils.unwrap(right).nodes
            if not values:
                raise StilusError(
                    '"in" given invalid right-hand operand, '
                    "expecting an expression"
                )

            # 'prop' in object
            if len(values) == 1 and values[0].node_name == "objectnode":
                return Boolean(values[0].has(self.hash()))

            for value in values:
                if str(value.hash()) == str(self.hash()):
                    return Boolean(True)

            return Boolean(False)
        elif op == "&&":
            a = self.to_boolean()
            b = right.to_boolean()
            if a.value is True and b.value is True:
                return right
            elif a.value is False:
                return self
            return right
        elif op == "[]":
            raise StilusError(f"cannot perform {self}[{right}]")
        else:
            raise StilusError(f"cannot perform {self} {op} {right}")

    def should_coerce(self, op: str) -> bool:
        """Return False if `op` is generally not coerced."""
        if op in ["is a", "in", "||", "&&"]:
            return False
        return True

    def coerce(self, other: Type["Node"]) -> Type["Node"]:
        """Default coercion raises."""
        if other.node_name == self.node_name:
            return other
        raise CoercionError(f"cannot coerce {other} to {self.node_name}")

    def clone(self, parent=None, node=None):
        """Return this node."""
        return self

    def to_json(self):
        """Return a JSON representation of this node."""
        return json.dumps([self.lineno, self.column, self.filename])



from pathlib import Path

from PIL import Image
from bs4 import BeautifulSoup

from stilus.exceptions import StilusError
from stilus.nodes.expression import Expression
from stilus.nodes.unit import Unit
from stilus.utils import assert_type, lookup


def _x_and_y_from_svg(path):
    with open(path) as f:
        soup = BeautifulSoup(f, "lxml")
    width = soup.svg["width"]
    height = soup.svg["height"]
    return [Unit(width), Unit(height)]


def image_size(img: str, ignore_errors=False, evaluator=None):
    assert_type(img, "string", "img")
    p = Path(img.string)
    path = lookup(p, evaluator.paths)
    if p.suffix == ".svg":
        return _x_and_y_from_svg(path)
    if path:
        with Image.open(path) as image:
            x, y = image.size
            expression = Expression()
            expression.nodes = [Unit(x, "px"), Unit(y, "px")]
            return expression
    elif ignore_errors:
        expression = Expression()
        expression.nodes = [Unit(0), Unit(0)]
        return expression
    else:
        raise StilusError("Could not find image.")



import json

from stilus.exceptions import StilusError
from stilus.nodes.boolean import Boolean
from stilus.nodes.ident import Ident
from stilus.nodes.null import null
from stilus.nodes.object_node import ObjectNode
from stilus.utils import assert_string, coerce, lookup, parse_string


# todo: this needs a cleaning
def json_function(path, local=None, name_prefix=None, evaluator=None):
    assert_string(path, "path")

    def convert(content, options):
        ret = ObjectNode()
        leave_strings = options.get("leave-strings").to_boolean()
        for key in content:
            val = content[key]
            if isinstance(val, dict):
                ret.set(key, convert(val, options))
            else:
                val = coerce(val, raw=False)
                if (
                    val
                    and val.node_name == "string"
                    and leave_strings.is_false()
                ):
                    val = parse_string(val.string)
                ret.set(key, val)
        return ret

    # lookup
    path = path.string
    found = lookup(path, evaluator.paths, evaluator.filename)
    options = None
    if (
        local
        and hasattr(local, "node_name")
        and local.node_name == "objectnode"
    ):
        options = local

    if not found:
        if options:
            optional = options.get("optional")
            if optional and optional.first().is_true():
                return null
        raise StilusError(f"failed to locate .json file {path}")

    # read
    with open(found, "r") as reader:
        content = json.load(reader)

    if options:
        return convert(content, options)
    else:
        old_json_function(content, local, name_prefix, evaluator)


def old_json_function(content, local=None, name_prefix=None, evaluator=None):
    def convert(content, prefix=None):
        if prefix:
            prefix = prefix + "-"
        else:
            prefix = ""
        for key in content:
            val = content[key]
            name = prefix + key
            if isinstance(val, dict):
                convert(val, name)
            else:
                val = coerce(val, raw=False)
                if val.node_name == "string":
                    val = parse_string(val.string)
                scope.add(Ident(name_prefix + name, val))

    if name_prefix:
        assert_string(name_prefix, "name_prefix")
        name_prefix = name_prefix.value
    else:
        name_prefix = ""

    if local:
        local = local.to_boolean()
    else:
        local = Boolean(local)

    if local.is_true():
        scope = evaluator.get_current_scope()
    else:
        scope = evaluator.common.scope()

    convert(content)



COMBINATORS = [">", "+", "~"]


# TODO: this class is badly migrated - needs work!
# TODO: this class needs tests!
# TODO: this class needs proper getter and setters annotations!
class SelectorParser:
    def __init__(self, string: str, stack=None, parts=None):
        self.string = string
        self.stack = stack
        if stack is None:
            self.stack = []
        self.parts = parts
        if parts is None:
            self.parts = []
        self.pos = 0
        self.level = 2
        self.nested = True
        self.ignore = False
        self.value = None
        self.raw = False

    def skip(self, len):
        self.string = self.string[len:]
        self.pos += len

    def skip_spaces(self):
        while " " == self.string[0]:
            self.skip(1)

    def advance(self):
        tok = self.root()
        if tok:
            return tok
        tok = self.relative()
        if tok:
            return tok
        tok = self.initial()
        if tok:
            return tok
        tok = self.escaped()
        if tok:
            return tok
        tok = self.parent()
        if tok:
            return tok
        tok = self.partial()
        if tok:
            return tok
        tok = self.char()
        if tok:
            return tok
        return None

    def root(self):
        r"""'/'"""
        if (
            not self.pos
            and "/" == self.string[0]
            and "deep" != self.string[1:5]
        ):
            self.nested = False
            self.skip(1)

    def relative(self, multi=None):
        r"""'../'"""
        if (not self.pos or multi) and "../" == self.string[0:3]:
            self.nested = False
            self.skip(3)
            while self.relative(True):
                self.level += 1
            if not self.raw:
                if len(self.stack) >= self.level:
                    return self.stack[len(self.stack) - self.level]
                else:
                    self.ignore = True

    def initial(self):
        r"""'~/'"""
        if not self.pos and "~/" == self.string[0:2]:
            self.nested = False
            self.skip(2)
            if self.stack and len(self.stack) > 0:
                return self.stack[0]
            else:
                return None

    def escaped(self):
        r"""'\' ('&' | '^')"""
        if self.string and len(self.string) > 0 and "\\" == self.string[0]:
            char = self.string[1]
            if char in ["&", "^"]:
                self.skip(2)
                return char

    # fixme: add the raw attribute/function check
    def parent(self):
        r"""'&"""
        if self.string and len(self.string) > 0 and "&" == self.string[0]:
            self.nested = False
            if not self.pos and (not self.stack or self.raw):
                i = 0
                for i, char in enumerate(self.string[1:], start=1):
                    if char != " ":
                        break
                if self.string[i] in COMBINATORS:
                    self.skip(i + 1)

            self.skip(1)
            if not self.raw:
                if self.stack and len(self.stack) > 0:
                    return self.stack[len(self.stack) - 1]
        return None

    def partial(self):
        r"""'^[' range ']'"""
        if "^[" == self.string[0:2]:
            self.skip(2)
            self.skip_spaces()
            ret = self.range()
            self.skip_spaces()
            if "]" != self.string[0]:
                return "^["
            self.nested = False
            self.skip(1)
            if ret:
                return ret
            else:
                self.ignore = True

    def number(self):
        r"""'-'? 0-9+"""
        i = 0
        ret = ""
        if "-" == self.string[i]:
            ret += self.string[i]
            i += 1

        while self.string[i].isdigit():
            ret += self.string[i]
            i += 1

        if ret:
            self.skip(i)
            return int(ret)

    def range(self):
        r"""number ('..' number)?"""
        start = self.number()
        ret = None

        if ".." == self.string[0:2]:
            self.skip(2)
            end = self.number()
            length = len(self.parts)

            if start < 0:
                start = length + start - 1
            if end < 0:
                end = length + end - 1

            if start > end:
                start, end = end, start

            def selector_value(selector):
                if selector["nested"]:
                    return f" {selector['value']}"
                else:
                    return f"{selector['value']}"

            def selector(part):
                selector = SelectorParser(part, self.stack, self.parts)
                selector.raw = True
                return selector.parse()

            if end < length - 1:
                ret = []
                for part in self.parts[start : end + 1]:
                    s = selector(part)
                    ret.append(selector_value(s))
                return "".join(ret).strip()

        else:
            # todo: fix me!
            if len(self.stack) > 0:
                if start < 0:
                    ret = self.stack[len(self.stack) + start - 1]
                else:
                    try:
                        ret = self.stack[start]
                    except IndexError:
                        ret = None
            else:
                ret = None
        if ret:
            return ret
        else:
            self.ignore = True

    def char(self):
        r""".+"""
        char = None
        if self.string:
            char = self.string[0]
        self.skip(1)
        return char

    def parse(self):
        r"""Parses the selector."""
        value = ""
        while self.string:
            next = self.advance()
            if next:
                value += next
            if self.ignore:
                value = ""
                break
        return {"value": value.rstrip(), "nested": self.nested}



import json

from .node import Node


class Namespace(Node):
    def __init__(self, value, prefix=None, lineno=1, column=1):
        super().__init__(value, lineno=lineno, column=column)
        self.prefix = prefix

    def __str__(self):
        prefix = f"{self.prefix} " if self.prefix else ""
        return f"@namespace {prefix}{self.value}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.value, self.prefix

    def __eq__(self, other):
        if isinstance(other, Namespace):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def to_json(self):
        return json.dumps(
            {
                "__type": "Charset",
                "val": self.value,
                "prefix": self.prefix,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class Literal(Node):
    def __init__(self, string, css=False, prefixed=False, lineno=1, column=1):
        super().__init__(string, lineno=lineno, column=column)
        self.string = string
        self.css = css
        self.prefixed = prefixed

    def __str__(self):
        return f"{self.value}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.string, self.prefixed

    def __hash__(self):
        return hash(self.__key())

    # checkme: really no need - just call super()?
    def hash(self):
        return self.value

    def __eq__(self, other):
        if isinstance(other, Literal):
            return self.__key() == other.__key()
        return False

    def to_json(self):
        return json.dumps(
            {
                "__type": "Literal",
                "val": self.value,
                "string": self.string,
                "prefixed": self.prefixed,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def coerce(self, other):
        if other.node_name in ["ident", "string", "literal"]:
            return Literal(other.string)
        else:
            super().coerce(other)

    def operate(self, op, right, value=None):
        if op == "+":
            return Literal(self.string + self.coerce(right.first()).string)
        else:
            return super().operate(op, right)



from stilus.nodes.ident import Ident
from stilus.nodes.string import String
from stilus.utils import assert_string, assert_type


def substr(value, start, length=None, evaluator=None):
    assert_string(value, "val")
    assert_type(start, "unit", "start")
    start = int(start.value)
    if length:
        length = start + int(length.value)
    result = value.string[start:length]
    if isinstance(value, Ident):
        return Ident(result)
    else:
        return String(result)



import json

from .node import Node


class Property(Node):
    def __init__(self, segments, expr=None, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.segments = segments
        self.expr = expr
        self.name = None
        self.literal = None

    def __str__(self):
        strings = [str(item) for item in self.segments]
        return f'property({"".join(strings)}, {self.expr})'

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.segments, self.expr

    def __eq__(self, other):
        if isinstance(other, Property):
            return self.__key() == other.__key()
        return False

    def clone(self, parent=None, clone=None):
        clone = Property(self.segments)
        clone.name = self.name
        if self.literal:
            clone.literal = self.literal
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        clone.segments = [node.clone(parent, clone) for node in self.segments]
        if self.expr:
            clone.expr = self.expr.clone(parent, clone)
        return clone

    def to_json(self):
        # FIXME: add expr and literal to this dict
        return json.dumps(
            {
                "__type": "Property",
                "segments": self.segments,
                "name": self.name,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def operate(self, op, right, value):
        return self.expr.operate(op, right, value)



import json

from .node import Node


class Comment(Node):
    def __init__(
        self, value: str, suppress: bool, inline: bool, lineno=1, column=1
    ):
        super().__init__(value, lineno=lineno, column=column)
        self.suppress = suppress
        self.inline = inline

    def __str__(self):
        return self.value

    def __repr__(self):
        return self.__str__()

    def to_json(self):
        return json.dumps(
            {
                "__type": "Comment",
                "str": self.value,
                "suppress": self.suppress,
                "inline": self.inline,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.utils import assert_string, assert_type


def prefix_classes(prefix, block=None, evaluator=None):
    assert_string(prefix, "prefix")
    assert_type(block, "block", "block")

    _prefix = evaluator.prefix

    evaluator.options["prefix"] = evaluator.prefix = prefix.string
    block = evaluator.visit(block)
    evaluator.options["prefix"] = evaluator.prefix = _prefix
    return block



import json

from .expression import Expression
from .node import Node


class ReturnNode(Node, Exception):
    def __init__(self, expression: Expression = None, lineno=1, column=1):
        Node.__init__(self, lineno=lineno, column=column)
        self.expression = expression

    def __str__(self):
        return f"Return[{self.expression}]"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.type, self.expression

    def __eq__(self, other):
        if isinstance(other, ReturnNode):
            return self.__key() == other.__key()
        return False

    def clone(self, parent=None, node=None):
        clone = ReturnNode(lineno=self.lineno, column=self.column)
        clone.filename = self.filename
        if self.expression:
            clone.expression = self.expression.clone(parent, clone)
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Return",
                "expr": self.expression,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



from stilus.nodes.null import null
from stilus.utils import unwrap


def shift(expr, *args, evaluator=None):
    expr = unwrap(expr)
    if len(expr.nodes) > 0:
        return expr.nodes.pop(0)
    else:
        return null



import json

from .node import Node


class Atblock(Node):
    def __init__(self, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.block = None

    def __str__(self):
        return f"@block"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.block

    def __eq__(self, other):
        if isinstance(other, Atblock):
            return self.__key() == other.__key()
        return False

    @property
    def nodes(self):
        if self.block:
            return self.block.nodes
        else:
            return []

    def clone(self, parent=None, node=None):
        clone = Atblock()
        clone.block = self.block.clone(parent, clone)
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Atblock",
                "block": self.block,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['stilus', 'stilus.functions', 'stilus.nodes', 'stilus.stack', 'stilus.visitor']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4',
 'click',
 'deprecated',
 'lxml',
 'pillow',
 'pyyaml',
 'setuptools>=44.0.0,<45.0.0',
 'watchdog>=0.8.3,<0.9.0',
 'wheel']

entry_points = \
{'console_scripts': ['stilus = stilus.cli:stilus']}

setup_kwargs = {
    'name': 'stilus',
    'version': '0.9.0',
    'description': 'A Stylus css compiler in Python.',
    'long_description': '\n# Stilus\n\n[![Travis (.org) branch](https://img.shields.io/travis/jw/stilus/master.svg?style=flat-square)](https://travis-ci.org/jw/stilus)\n[![Codecov branch](https://img.shields.io/codecov/c/github/jw/stilus/master.svg?style=flat-square)](https://codecov.io/gh/jw/stilus)\n[![PyPI - Downloads](https://img.shields.io/pypi/dm/stilus.svg?style=flat-square)](https://pypi.org/project/stilus/#files)\n[![PyPI](https://img.shields.io/pypi/v/stilus.svg?style=flat-square)](https://pypi.org/project/stilus/#history)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/stilus.svg?style=flat-square)](https://pypi.org/project/stilus/#description)\n[![PyPI - License](https://img.shields.io/pypi/l/stilus?style=flat-square)](https://pypi.org/project/stilus/)\n\nA Stylus css compiler in Python.\n\n## Installation\n\nInstall and update using [pip](https://pypi.org/project/pip/):\n\n    $ python -m pip install -U stilus\n\nor \n\n    $ pip install -U stilus\n\n## Documentation\n\nThere is [some documentation](https://stilus.readthedocs.io) available.\n\n## Missing\n\nSome parts of Stylus are not in Stilus yet:\n\n - Caching\n - Plugins\n - Documentation\n\nAll help is appreciated!\n',
    'author': 'Jan Willems',
    'author_email': 'jw@elevenbits.com',
    'maintainer': 'Jan Willems',
    'maintainer_email': 'jw@elevenbits.com',
    'url': 'https://github.com/jw/stilus',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import json

from .boolean import Boolean
from .expression import Expression
from .node import Node

FACTOR_TABLE = {
    "mm": {"value": 1, "label": "mm"},
    "cm": {"value": 10, "label": "mm"},
    "in": {"value": 25.4, "label": "mm"},
    "pt": {"value": 25.4 / 72, "label": "mm"},
    "ms": {"value": 1, "label": "ms"},
    "s": {"value": 1000, "label": "ms"},
    "Hz": {"value": 1, "label": "Hz"},
    "kHz": {"value": 1000, "label": "Hz"},
}


# todo: remove this function!
def is_weird_float(value):
    if isinstance(value, float) and f"{value}".endswith(".0"):
        return True
    return False


def has_many_zeros_and_a_number(value):
    if isinstance(value, (float, str)):
        s = f"{value}"
        if "." in s and len(s.split(".")[1]) > 10 and s[-1].isdigit():
            return True
    return False


class Unit(Node):
    def __init__(self, value, type=None, lineno=1, column=1):
        """
        Initialize a new `Unit` with the given `val` and unit `type` such as
        "px", "pt", "in", etc.
        :param value:
        :param type:
        """
        super().__init__(value, lineno=lineno, column=column)
        self.type = type

    def __str__(self):
        type = self.type if self.type else ""
        # remove trailing zeros
        if is_weird_float(self.value):  # fixme: do not use this!
            v = f"{self.value}".rstrip("0").rstrip(".")
        elif has_many_zeros_and_a_number(self.value):  # fixme: remove this!
            v = f"{self.value[:-1]}".rstrip("0")
        else:
            v = self.value
        return f"{v}{type}"

    def __repr__(self):
        return self.__str__()

    def to_boolean(self):
        return Boolean(True) if self.type else Boolean(self.value)

    def to_json(self):
        return json.dumps(
            {
                "__type": "Unit",
                "value": self.value,
                "type": self.type,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )

    def clone(self, parent=None, node=None):
        clone = Unit(self.value, self.type)
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def hash(self):
        if isinstance(self.value, int) or is_weird_float(self.value):  # fixme!
            return int(self.value)
        else:
            return self.value

    def operate(self, op, right: Node, value=None):  # noqa: C901
        type = None
        if self.type:
            type = self.type
        elif hasattr(right.first(), "type") and right.first().type:
            type = right.first().type

        # swap color
        if right.node_name in ["rgba", "hsla"]:
            return right.operate(op, self)

        # operate
        if self.should_coerce(op):
            right = right.first()

            if (
                hasattr(self, "type")
                and self.type != "%"
                and op in ["-", "+"]
                and hasattr(right, "type")
                and right.type == "%"
            ):
                right = Unit(self.value * (right.value / 100), "%")
            else:
                right = self.coerce(right)

            if op == "-":
                return Unit(self.value - right.value, type)
            elif op == "+":
                # keyframes interpolation
                if type:
                    type = type
                else:
                    type = right.type if right.type == "%" else None
                return Unit(self.value + right.value, type)
            elif op == "/":
                return Unit(self.value / right.value, type)
            elif op == "*":
                return Unit(self.value * right.value, type)
            elif op == "%":
                return Unit(self.value % right.value, type)
            elif op == "**":
                return Unit(self.value ** right.value, type)
            elif op in ["..", "..."]:
                start = int(self.value)
                end = int(right.value)
                expr = Expression()
                inclusive = ".." == op
                if start < end:
                    while True:
                        expr.append(Unit(start))
                        start += 1
                        if inclusive and start > end:
                            break
                        if not inclusive and start >= end:
                            break
                else:
                    while True:
                        expr.append(Unit(start))
                        start -= 1
                        if inclusive and start < end:
                            break
                        if not inclusive and start <= end:
                            break
                return expr

        return super().operate(op, right)

    def coerce(self, other):
        if other.node_name == "unit":
            a = self
            b = other

            if a.type:
                factor_a = FACTOR_TABLE.get(
                    a.type, {"value": 1, "label": a.type}
                )
            else:
                factor_a = None
            if b.type:
                factor_b = FACTOR_TABLE.get(
                    b.type, {"value": 1, "label": b.type}
                )
            else:
                factor_b = None

            if (
                factor_a
                and factor_b
                and factor_a["label"] == factor_b["label"]
            ):
                b_value = b.value * (factor_b["value"] / factor_a["value"])
                return Unit(b_value, a.type)
            else:
                return Unit(b.value, a.type)
        elif other.node_name == "string":
            # keyframes interpretation
            if other.value == "%":
                return Unit(0, "%")
            try:
                value = float(other.value.split()[0])
            except ValueError:
                return super().coerce(other)
            return Unit(value)
        else:
            return super().coerce(other)



import re

from stilus import utils
from stilus.utils import assert_type


def get_flags(flags=None):
    f = 0
    if flags:
        for c in flags.string:
            if "i" == c:
                f |= re.I
            if "m" == c:
                f |= re.M
    return f


def stilus_findall(pattern, string, flags=None):
    p = re.compile(pattern, get_flags(flags))
    if flags and "g" in flags.string:
        return p.findall(string)
    else:
        m = p.search(string)
        if m:
            return [m.group(0)] + list(m.groups())
    from stilus.nodes.null import null

    return null


def match(pattern, val, flags=None, evaluator=None):
    assert_type(pattern, "string", "pattern")
    utils.assert_string(val, "val")
    return stilus_findall(pattern.value, val.string, flags)



import math as m
from stilus.nodes.unit import Unit
from stilus.utils import assert_type, stilus_round


def tan(angle, evaluator=None):
    assert_type(angle, "unit", "angle")
    radians = angle.value
    if angle.type == "deg":
        radians = m.radians(radians)
    # start of Stylus tan generation...
    pow = m.pow(10, 9)
    sin = float(stilus_round(m.sin(radians) * pow)) / pow
    cos = float(stilus_round(m.cos(radians) * pow)) / pow
    if cos == 0:
        tan = "Infinity"
    else:
        tan = float(stilus_round(pow * sin / cos)) / pow
    if tan == 0:
        tan = 0
    # ...which in fact should be m.tan(radians)
    return Unit(tan, "")



import json

from .node import Node


class Import(Node):
    def __init__(self, expr=None, once=False, lineno=1, column=1):
        super().__init__(lineno, column)
        self.path = expr
        self.once = once
        self.mtime = None

    def __str__(self):
        return f"import({self.path}, {self.once}"

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.__key())

    def __key(self):
        return self.path, self.once

    def __eq__(self, other):
        if isinstance(other, Import):
            return self.__key() == other.__key()
        return False

    def clone(self, parent=None, node=None):
        clone = Import(lineno=self.lineno, column=self.column)
        clone.path = self.path.clone(parent, clone)
        clone.once = self.once
        clone.mtime = self.mtime
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "Import",
                "path": self.path,
                "once": self.once,
                "mtime": self.mtime,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import json

from .node import Node


class If(Node):
    def __init__(self, cond=None, negate=None, lineno=1, column=1):
        super().__init__(lineno=lineno, column=column)
        self.cond = cond
        self.elses = []
        self.postfix = None
        self.block = None
        self.negate = None
        if negate and isinstance(negate, Node):
            self.block = negate
        else:
            self.negate = negate

    def __str__(self):
        if self.elses:
            else_string = f' else ({", ".join(self.elses)})'
        else:
            else_string = " [no else]"
        return f"if {self.cond}{else_string}"

    def __repr__(self):
        return self.__str__()

    def __key(self):
        return self.node_name, self.cond

    def __eq__(self, other):
        if isinstance(other, If):
            return self.__key() == other.__key()
        return False

    def __hash__(self):
        return hash(self.__key())

    def clone(self, parent=None, node=None):
        clone = If()
        clone.cond = self.cond.clone(parent)
        clone.block = self.block.clone(parent, clone)
        clone.elses = [node.clone(parent, clone) for node in self.elses]
        clone.negate = self.negate
        clone.postfix = self.postfix
        clone.lineno = self.lineno
        clone.column = self.column
        clone.filename = self.filename
        return clone

    def to_json(self):
        return json.dumps(
            {
                "__type": "If",
                "cond": self.cond,
                "block": self.block,
                "elses": self.elses,
                "negate": self.negate,
                "postfix": self.postfix,
                "lineno": self.lineno,
                "column": self.column,
                "filename": self.filename,
            }
        )



import re
from collections import deque

from .nodes.boolean import Boolean
from .nodes.color import RGBA
from .nodes.null import null
from .nodes.string import String
from .nodes.unit import Unit
from .nodes.comment import Comment
from .nodes.ident import Ident
from .nodes.literal import Literal


class Token:
    def __init__(self, type, value=None, space=None, lineno=None, column=None):
        """A token."""
        self.type = type
        self.value = value
        self.space = space
        self.anonymous = False
        self.lineno = lineno
        self.column = column

    def __str__(self):
        """Return the value if available, the tokens type otherwise."""
        if self.value:
            return f"{self.value}"
        else:
            return f"{self.type}"

    def __repr__(self):
        """Return the value if available, the tokens type otherwise."""
        return str(self)

    def __key(self):
        return (
            self.type,
            self.value,
            self.space,
            self.anonymous,
            self.lineno,
            self.column,
        )

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, Token):
            return self.__key() == other.__key()
        return False


class Lexer:
    alias = {
        "and": "&&",
        "or": "||",
        "is": "==",
        "isnt": "!=",
        "is not": "!=",
        ":=": "?=",
    }

    def __init__(self, s: str, options: dict):
        self.stash = deque([])
        self.indent_stack = deque([])
        self.indent_re = None
        self.lineno = 1
        self.column = 1
        self.options = options

        self.prev = None
        self.is_url = False
        self.at_eos = False

        self.s = self.clean(s)

    def __eq__(self, other):
        if isinstance(other, Lexer):
            return self.s == other.s and self.options == other.options
        return False

    def clean(self, s: str) -> str:
        # handle UTF-8 BOM
        s = s[1:] if len(s) > 1 and "\uFEFF" == s[0] else s

        s = re.sub(r"\s+$", "\n", s)
        s = re.sub(r"\r\n?", "\n", s)
        s = re.sub(r"\\ *\n", "\r", s)

        def _comment(match):
            # TODO: this needs a rewrite
            m = match.group(0)
            value = match.groups()[0]
            offset = match.start(0)
            string = match.string
            in_comment = string.rfind("/*", 0, offset) > string.rfind(
                "*/", 0, offset
            )
            comment_index = string.rfind("//", 0, offset)
            i = s.rfind("\n", 0, offset)
            double = 0
            single = 0

            if comment_index != -1 and comment_index > i:
                while i != offset:
                    if "'" == s[i]:
                        single = single - 1 if single else single + 1
                    if '"' == s[i]:
                        double = double - 1 if double else double + 1
                    if "/" == s[i] and "/" == s[i + 1]:
                        in_comment = not single and not double
                        break
                    i += 1

            return m if in_comment else value + "\r"

        s = re.sub(
            r"([,(:](?!\/\/[^ ])) *(?:\/\/[^\n]*|\/\*.*?\*\/)?\n\s*",
            _comment,
            s,
        )
        s = re.sub(r"\s*\n[ \t]*([,)])", _comment, s)

        return s

    def _skip_number(self, len):
        self.s = self.s[len:]
        self.column += len

    def _skip_string(self, str):
        self.s = self.s[len(str) :]
        self.move(str)

    def move(self, str):
        lines = str.count("\n")
        self.lineno += lines
        idx = str.rfind("\n")
        if idx == -1:
            self.column += len(str)
        else:
            self.column = len(str) - idx

    def is_part_of_selector(self):
        tok = self.stash[-1] if self.stash else self.prev
        if tok and tok.type == "color":
            return 2 == len(tok.value.raw)
        elif tok and (tok.type in [".", "["]):
            return True
        else:
            return False

    def next(self) -> Token:
        tok = self.stash.popleft() if len(self.stash) > 0 else self.advance()
        self.prev = tok
        self.at_eos = tok.type == "eos"
        return tok

    def peek(self):
        return self.lookahead(1)

    def lookahead(self, n):
        fetch = n - len(self.stash)
        while fetch > 0:
            fetch -= 1
            self.stash.append(self.advance())
        n -= 1
        return self.stash[n]

    def __next__(self) -> Token:
        if self.at_eos:
            raise StopIteration
        return self.next()

    def __iter__(self):
        return self

    # todo: use a dict to clean this up?
    def advance(self):  # noqa: C901
        lineno = self.lineno
        column = self.column
        tok = self.eos()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.null()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.sep()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.keyword()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.urlchars()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.comment()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.newline()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.escaped()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.important()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.literal()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.anonymous_function()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.atrule()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.function()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.brace()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.paren()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.color()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.string()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.unit()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.namedop()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.boolean()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.unicode()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.ident()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.op()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.eol()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.space()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        tok = self.selector()
        if tok:
            return self._tok_with_location(tok, lineno, column)
        return None

    def eos(self):
        """
        eos | trailing outdents
        """
        if self.s != "":
            return False
        if self.indent_stack:
            self.indent_stack.popleft()
            return Token("outdent")
        else:
            return Token("eos")

    def null(self):
        """
        null
        """
        match = re.match(r"^(null)\b[ \t]*", self.s)
        if match:
            self._skip_string(match.group())
            if self.is_part_of_selector():
                return Token("ident", Ident(match.group(0)))
            else:
                return Token("null", value=null)

    def newline(self):
        """
        '\n' ' '+
        """
        if self.indent_re:
            match = re.match(self.indent_re, self.s)

        # figure out if we are using tabs or spaces
        else:
            # try tabs first
            possible_re = r"^\n([\t]*)[ \t]*"
            match = re.match(possible_re, self.s)
            if match and not match.group(1):
                # try spaces next
                possible_re = r"^\n([ \t]*)"
                match = re.match(possible_re, self.s)
            # established
            if match and match.group(1):
                self.indent_re = possible_re

        if match:
            indents = len(match.group(1))

            self._skip_string(match.group(0))

            if self.s and self.s[0] in [" ", "\t"]:
                raise SyntaxError(
                    "Invalid indentation. You can use tabs or "
                    "spaces to indent, but not both."
                )

            # blank line
            if self.s and self.s[0] == "\n":
                return self.advance()

            # outdent
            if self.indent_stack and indents < self.indent_stack[0]:
                while self.indent_stack and self.indent_stack[0] > indents:
                    self.stash.append(Token("outdent"))
                    self.indent_stack.popleft()
                tok = self.stash.pop()

            # indent
            elif indents and (
                len(self.indent_stack) == 0 or indents != self.indent_stack[0]
            ):
                self.indent_stack.appendleft(indents)
                tok = Token("indent")

            # newline
            else:
                tok = Token("newline")

            return tok

        return False

    def selector(self):
        """
        ^|[^\n,;]+
        """
        match = re.match(r"^\^|.*?(?=\/\/(?![^\[]*\])|[,\n{])", self.s)
        if match and match.group(0):
            selector = match.group(0)
            self._skip_string(selector)
            return Token("selector", selector)

    def space(self):
        """
        ' '+ | '\t'+
        """
        match = re.match(r"^([ \t]+)", self.s)
        if match:
            self._skip_string(match.group(0))
            return Token("space")

    def eol(self):
        """
        '\r'
        """
        if self.s and self.s[0] == "\r":
            self.lineno += 1
            self._skip_number(1)
            return self.advance()

    def op(self):
        """
        ',' | '+' | '+=' | '-' |  '-=' |  '*' | '*=' | '/' | '/=' | '%' |
        '%=' | '**' |  '!' | '&' | '&&' | '||' | '>' | '>=' | '<' | '<=' |
        '=' | '==' | '!=' | '!' | '~' | '?=' | ':=' | '?' | ':' | '[' |
        ']' | '.' | '..' |'...'
        """
        match = re.match(
            r"^([.]{1,3}|&&|\|\||[!<>=?:]=|\*\*|[-+*\/%]=?|"
            r"[,=?:!~<>&\[\]])([ \t]*)",
            self.s,
        )
        if match:
            op = match.group(1)
            self._skip_string(match.group(0))
            op = self.alias.get(op, op)
            tok = Token(op, op)
            tok.space = match.group(2)
            self.is_url = False
            return tok

    def sep(self):
        """
        ';' [ \t]*
        """
        match = re.match(r"^;[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            return Token(";")

    def keyword(self):
        """
        'if' | 'else' | 'unless' | 'return' | 'for' | 'in'
        """
        match = re.match(r"^(return|if|else|unless|for|in)\b[ \t]*", self.s)
        if match:
            keyword = match.group(1)
            self._skip_string(match.group(0))
            if self.is_part_of_selector():
                tok = Token("ident", Ident(match.group(0)))
            else:
                tok = Token(keyword, keyword)
            return tok

    def urlchars(self):
        """
        url char
        """
        if not self.is_url:
            return
        match = re.match(r"^[\/:@.;?&=*!,<>#%0-9]+", self.s)
        if match:
            self._skip_string(match.group(0))
            return Token("literal", Literal(match.group(0)))

    def comment(self):
        """
        '//' *
        """
        # single line
        if (
            self.s
            and len(self.s) > 2
            and "/" == self.s[0]
            and "/" == self.s[1]
        ):
            end = self.s.find("\n")
            if -1 == end:
                end = len(self.s)
            self._skip_number(end)
            return self.advance()

        # multi-line
        if (
            self.s
            and len(self.s) > 2
            and "/" == self.s[0]
            and "*" == self.s[1]
        ):
            end = self.s.find("*/")
            suppress = True
            inline = False
            s = self.s[0 : end + 2]
            lines = len(re.split(r"[\n|\r]", s)) - 1
            self.lineno += lines
            self._skip_number(end + 2)
            # output
            if "!" == s[2]:
                s = s.replace("*!", "*")
                suppress = False
            if self.prev and ";" == self.prev.type:
                inline = True
            return Token("comment", Comment(s, suppress, inline))

    def escaped(self):
        """
        '\\' . ' '*
        """
        match = re.match(r"^\\(.)[ \t]*", self.s)
        if match:
            escape = match.group(1)
            self._skip_string(match.group(0))
            if match.group(0).startswith("\\."):
                escape = "\\."
            return Token("ident", Literal(escape))

    def important(self):
        """
        '!important' ' '*
        """
        match = re.match(r"^!important[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            return Token("ident", Literal("!important"))

    def literal(self):
        """
        '@css' ' '* '{' .* '}' ' '*
        """
        match = re.match(r"^@css[ \t]*{", self.s)
        if match:
            self._skip_string(match.group(0))
            braces = 1
            css = ""
            while True:
                c = self.s[0]
                self.s = self.s[1:]
                if c == "{":
                    braces += 1
                elif c == "}":
                    braces -= 1
                elif c in ["\n", "\r"]:
                    self.lineno += 1
                css += c
                if not braces:
                    break
            css = re.sub(r"\s*}$", "", css)
            return Token("literal", Literal(css, css=True))

    def anonymous_function(self):
        """
        '@('
        """
        if "@" == self.s[0] and "(" == self.s[1]:
            self._skip_number(2)
            tok = Token("function", Ident("anonymous"))
            tok.anonymous = True
            return tok

    def atrule(self):
        r"""
        # '@' (-(\w+)-)?[a-zA-Z0-9-_]+
        """
        match = re.match(r"^@(?:-(\w+)-)?([a-zA-Z0-9-_]+)[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            vendor = match.group(1)
            type = match.group(2)
            if type in [
                "require",
                "import",
                "charset",
                "namespace",
                "media",
                "scope",
                "supports",
            ]:
                return Token(type)
            elif type == "document":
                return Token("-moz-document")
            elif type == "block":
                return Token("atblock")
            elif type in ["extend", "extends"]:
                return Token("extend")
            elif type == "keyframes":
                return Token(type, vendor)
            else:
                return Token("atrule", f"-{vendor}-{type}" if vendor else type)

    def function(self):
        r"""
         -*[_a-zA-Z$] [-\w\d$]* '('
        """
        match = re.match(r"^(-*[_a-zA-Z$][-\w\d$]*)\(([ \t]*)", self.s)
        if match:
            name = match.group(1)
            self._skip_string(match.group(0))
            self.is_url = "url" == name
            tok = Token("function", Ident(name))
            tok.space = match.group(2)
            return tok

    def brace(self):
        """
        '{' | '}'
        """
        match = re.match(r"^([{}])", self.s)
        if match:
            self._skip_number(1)
            brace = match.group(1)
            return Token(brace, brace)

    def paren(self):
        """
        '(' | ')' ' '*
        """
        match = re.match(r"^([()])([ \t]*)", self.s)
        if match:
            paren = match.group(1)
            self._skip_string(match.group(0))
            if ")" == paren:
                self.is_url = False
            tok = Token(paren, paren)
            tok.space = match.group(2)
            return tok

    def rrggbbaa(self):
        """
        #rrggbbaa
        """
        match = re.match(r"^#([a-fA-F0-9]{8})[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            rgb = match.group(1)
            r = int(rgb[0:2], 16)
            g = int(rgb[2:4], 16)
            b = int(rgb[4:6], 16)
            a = int(rgb[6:8], 16)
            color = RGBA(r, g, b, a / 255)
            color.raw = match.group(0)
            return Token("color", color)

    def rrggbb(self):
        """
        #rrggbb
        """
        match = re.match(r"^#([a-fA-F0-9]{6})[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            rgb = match.group(1)
            r = int(rgb[0:2], 16)
            g = int(rgb[2:4], 16)
            b = int(rgb[4:6], 16)
            color = RGBA(r, g, b, 1)
            color.raw = match.group(0)
            return Token("color", color)

    def rgba(self):
        """
        #rgba
        """
        match = re.match(r"^#([a-fA-F0-9]{4})[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            rgb = match.group(1)
            r = int(f"{rgb[0]}{rgb[0]}", 16)
            g = int(f"{rgb[1]}{rgb[1]}", 16)
            b = int(f"{rgb[2]}{rgb[2]}", 16)
            a = int(f"{rgb[3]}{rgb[3]}", 16)
            color = RGBA(r, g, b, a / 255)
            color.raw = match.group(0)
            return Token("color", color)

    def rgb(self):
        """
        #rgb
        """
        match = re.match(r"^#([a-fA-F0-9]{3})[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            rgb = match.group(1)
            r = int(f"{rgb[0]}{rgb[0]}", 16)
            g = int(f"{rgb[1]}{rgb[1]}", 16)
            b = int(f"{rgb[2]}{rgb[2]}", 16)
            color = RGBA(r, g, b, 1)
            color.raw = match.group(0)
            return Token("color", color)

    def nn(self):
        """
        #nn
        """
        match = re.match(r"^#([a-fA-F0-9]{2})[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            n = int(match.group(1), 16)
            color = RGBA(n, n, n, 1)
            color.raw = match.group(0)
            return Token("color", color)

    def n(self):
        """
        #n
        """
        match = re.match(r"^#([a-fA-F0-9]{1})[ \t]*", self.s)
        if match:
            self._skip_string(match.group(0))
            n = int(f"{match.group(1)}{match.group(1)}", 16)
            color = RGBA(n, n, n, 1)
            color.raw = match.group(0)
            return Token("color", color)

    def color(self):
        tok = self.rrggbbaa()
        if tok:
            return tok
        tok = self.rrggbb()
        if tok:
            return tok
        tok = self.rgba()
        if tok:
            return tok
        tok = self.rgb()
        if tok:
            return tok
        tok = self.nn()
        if tok:
            return tok
        tok = self.n()
        if tok:
            return tok

    def string(self):
        """
        '"' [^"]+ '"' | "'"" [^']+ "'"
        """
        match = re.match(r'^("[^"]*"|\'[^\']*\')[ \t]*', self.s)
        if match:
            self._skip_string(match.group(0))
            string = match.group(0).strip()[1:-1]
            quote = match.group(0)[0]
            return Token("string", String(string, quote))

    def unit(self):
        """
        '-'? (digit+ | digit* '.' digit+) unit
        """
        match = re.match(
            r"^(-)?(\d+\.\d+|\d+|\.\d+)(%|[a-zA-Z]+)?[ \t]*", self.s
        )
        if match:
            self._skip_string(match.group(0))
            n = float(match.group(2))
            if match.group(1) == "-":
                n = -n
            unit = Unit(n, match.group(3))
            unit.raw = match.group(0)
            return Token("unit", unit)

    def namedop(self):
        """
        'not' | 'and' | 'or' | 'is' | 'is not' | 'isnt' | 'is a' | 'is defined'
        """
        match = re.match(
            r"^(not|and|or|is a|is defined|isnt|is not|is)" r"(?!-)\b([ \t]*)",
            self.s,
        )
        if match:
            self._skip_string(match.group(0))
            op = match.group(1)
            if self.is_part_of_selector():
                tok = Token("ident", Ident(match.group(0)))
            else:
                op = self.alias.get(op, op)
                tok = Token(op, op)
            tok.space = match.group(2)
            return tok

    def boolean(self):
        """
        true | false
        """
        match = re.match(r"^(true|false)\b([ \t]*)", self.s)
        if match:
            self._skip_string(match.group(0))
            val = Boolean("true" == match.group(1))
            tok = Token("boolean", val)
            tok.space = match.group(2)
            return tok

    def unicode(self):
        """
        'U+' [0-9A-Fa-f?]{1,6}(?:-[0-9A-Fa-f]{1,6})?
        """
        match = re.match(
            r"^u\+[0-9a-f?]{1,6}(?:-[0-9a-f]{1,6})?",
            self.s,
            flags=re.IGNORECASE,
        )
        if match:
            self._skip_string(match.group(0))
            return Token("literal", Literal(match.group(0)))

    def ident(self):
        r"""
        -*[_a-zA-Z$] [-\w\d$]*
        """
        match = re.match(r"^-*[_a-zA-Z$][-\w\d$]*", self.s)
        if match:
            self._skip_string(match.group(0))
            return Token("ident", Ident(match.group(0)))

    def _tok_with_location(self, tok, lineno, column):
        tok.lineno = lineno
        tok.column = column
        return tok



from stilus.nodes.unit import Unit
from stilus.functions.hsla import hsla
from stilus.utils import assert_color


def hsl(hue, saturation=None, lightness=None, evaluator=None):
    if hue and (saturation is lightness is None):
        assert_color(hue, "color")
        return hue.hsla()
    return hsla(hue, saturation, lightness, Unit(1), evaluator)



from stilus.nodes.property import Property
from stilus.utils import assert_type, unwrap, assert_string


def add_property(name, expr, evaluator=None):
    assert_type(name, "expression", "name")
    name = unwrap(name).first()
    assert_string(name, "name")
    assert_type(expr, "expression", "expr")
    prop = Property([name], expr)
    block = evaluator.closest_block()
    head = block.nodes[0 : block.index]
    tail = block.nodes[block.index : len(block.nodes)]
    block.index += 1
    block.mixin = True  # this is a dirty hack
    head.append(prop)
    head.extend(tail)
    block.nodes = head
    return prop



import logging
from collections import deque

from .exceptions import ParseError
from .lexer import Lexer, Token
from .nodes.arguments import Arguments
from .nodes.atblock import Atblock
from .nodes.atrule import Atrule
from .nodes.binop import BinOp
from .nodes.block import Block
from .nodes.boolean import true
from .nodes.call import Call
from .nodes.charset import Charset
from .nodes.each import Each
from .nodes.expression import Expression
from .nodes.extend import Extend
from .nodes.feature import Feature
from .nodes.function import Function
from .nodes.group import Group
from .nodes.ident import Ident
from .nodes.ifnode import If
from .nodes.import_node import Import
from .nodes.keyframes import Keyframes
from .nodes.literal import Literal
from .nodes.media import Media
from .nodes.member import Member
from .nodes.namespace import Namespace
from .nodes.node import Node
from .nodes.nothing import Nothing
from .nodes.null import null
from .nodes.object_node import ObjectNode
from .nodes.params import Params
from .nodes.property import Property
from .nodes.query import Query
from .nodes.query_list import QueryList
from .nodes.return_node import ReturnNode
from .nodes.root import Root
from .nodes.selector import Selector
from .nodes.supports import Supports
from .nodes.ternary import Ternary
from .nodes.unaryop import UnaryOp
from .units import units

log = logging.getLogger(__name__)


class Parser:
    def __init__(self, s, options: dict):
        self.cond = None
        self.s = s
        self.options = options
        self.lexer = Lexer(s, options)
        self.prefix = options.get("prefix", "")
        self.root = options.get("root", Root())
        self.state = ["root"]
        self.stash = []
        self.parens = 0
        self.css = 0
        self.parent = None
        self.selector_scope = None
        self.bracketed = None
        self.in_property = None
        self._ident = None
        self.operand = None
        self.allow_postfix = None
        self.prev_state = None
        self.lineno = 1
        self.column = 1

    #
    # Selector composite tokens.
    #
    selector_tokens = [
        "ident",
        "string",
        "selector",
        "function",
        "comment",
        "boolean",
        "space",
        "color",
        "unit",
        "for",
        "in",
        "[",
        "]",
        "(",
        ")",
        "+",
        "-",
        "*",
        "*=",
        "<",
        ">",
        "=",
        ":",
        "&",
        "&&",
        "~",
        "{",
        "}",
        ".",
        "..",
        "/",
    ]

    #
    # CSS pseudo-classes and pseudo-elements.
    # See http://dev.w3.org/csswg/selectors4/
    #
    pseudo_selectors = [
        # Logical Combinations
        "matches",
        "not",
        # Linguistic Pseudo-classes
        "dir",
        "lang",
        # Location Pseudo-classes
        "any-link",
        "link",
        "visited",
        "local-link",
        "target",
        "scope",
        # User Action Pseudo-classes
        "hover",
        "active",
        "focus",
        "drop",
        # Time-dimensional Pseudo-classes
        "current",
        "past",
        "future",
        # The Input Pseudo-classes
        "enabled",
        "disabled",
        "read-only",
        "read-write",
        "placeholder-shown",
        "checked",
        "indeterminate",
        "valid",
        "invalid",
        "in-range",
        "out-of-range",
        "required",
        "optional",
        "user-error",
        # Tree-Structural pseudo-classes
        "root",
        "empty",
        "blank",
        "nth-child",
        "nth-last-child",
        "first-child",
        "last-child",
        "only-child",
        "nth-of-type",
        "nth-last-of-type",
        "first-of-type",
        "last-of-type",
        "only-of-type",
        "nth-match",
        "nth-last-match",
        # Grid-Structural Selectors
        "nth-column",
        "nth-last-column",
        # Pseudo-elements
        "first-line",
        "first-letter",
        "before",
        "after",
        # Non-standard
        "selection",
    ]

    def current_state(self):
        return self.state[-1]

    def previous_state(self):
        try:
            return self.state[-2]
        except Exception as e:
            # todo: handle this properly
            raise e

    def parse(self) -> Node:
        """
        Parse the given source string.

        :return: The root Node.
        :rtype: Node
        """
        block = self.parent = self.root
        while "eos" != self.peek().type:
            self.skip_whitespace()
            if "eos" == self.peek().type:
                break
            stmt = self.statement()
            self.accept(";")
            if stmt is None:
                self.error(
                    "unexpected token {peek}, not " "allowed at the root level"
                )
            block.append(stmt)
        return block

    def error(self, message: str):
        t = self.peek().type
        value = ""
        if self.peek().value:
            value = str(self.peek())
        if value.strip() == t.strip():
            value = ""
        raise ParseError(message.format(peek='"{}{}"'.format(t, value)))

    def accept(self, types):
        if isinstance(types, str):
            if self.peek().type == types:
                return self.next()
        elif isinstance(types, list):
            if self.peek().type in types:
                return self.next()
        return None

    def expect(self, type):
        if type != self.peek().type:
            self.error(f'expected "{type}", got {{peek}}')
        return self.next()

    def next(self) -> Token:
        if self.stash:
            tok = self.stash.pop()
        else:
            tok = self.lexer.next()
        if tok.value and isinstance(tok.value, Node):
            tok.value.lineno = tok.lineno
            tok.value.column = tok.column
        self.lineno = tok.lineno
        self.column = tok.column
        log.debug(f"Returning {tok}.")
        return tok

    def __iter__(self):
        return self

    def __next__(self):
        tok = self.next()
        if tok.type == "eos":
            raise StopIteration
        return tok

    def peek(self) -> Token:
        return self.lexer.peek()

    def lookahead(self, n):
        return self.lexer.lookahead(n)

    def is_selector_token(self, n: int) -> bool:
        """
        Check if the token at n is a valid selector token.
        :param n:
        :return:
        """
        la = self.lookahead(n).type
        if la == "for":
            return self.bracketed
        elif la == "[":
            self.bracketed = True
            return True
        elif la == "]":
            self.bracketed = False
            return True
        else:
            return la in self.selector_tokens

    def is_pseudo_selector(self, n):
        value = self.lookahead(n).value
        if value and hasattr(value, "name"):
            return value.name in self.pseudo_selectors
        return False

    def line_contains(self, type):
        i = 1
        la = self.lookahead(i)
        while True:
            if la.type in ["indent", "outdent", "newline", "eos"]:
                return False
            if la.type == type:
                return True
            i += 1
            la = self.lookahead(i)

    def selector_token(self):
        """Valid selector tokens."""
        if self.is_selector_token(1):
            if "{" == self.peek().type:

                if not self.line_contains("}"):
                    # unclosed - must be a block
                    return None

                # check if ':' is within the braces.
                # though not required by Stylus, chances
                # are if someone is using {} they will
                # use CSS-style props, helping us with
                # the ambiguity in this case
                i = 1
                la = self.lookahead(i)
                while la:
                    if la.type == "}":
                        # check empty block
                        if i == 2 or (
                            i == 3 and self.lookahead(i - 1).type == "space"
                        ):
                            return None
                        break
                    if la.type == ":":
                        return None
                    i += 1
                    la = self.lookahead(i)

            return self.next()
        return None

    def skip(self, tokens):
        while self.peek().type in tokens:
            self.next()

    def skip_whitespace(self):
        self.skip(["space", "indent", "outdent", "newline"])

    def skip_spaces(self):
        self.skip(["space"])

    def skip_spaces_and_comments(self):
        self.skip(["space", "comment"])

    def looks_like_function_definition(self, i):
        return self.lookahead(i).type in ["indent", "{"]

    def looks_like_selector(self, from_property=False):  # noqa: C901
        i = 1
        brace = None

        # real property
        if (
            from_property
            and ":" == self.lookahead(i + 1).type
            and (
                self.lookahead(i + 1).space
                or "indent" == self.lookahead(i + 2).type
            )
        ):
            return False

        # assume selector when an ident is followed by a selector
        while "ident" == self.lookahead(i).type and (
            "newline" == self.lookahead(i + 1).type
            or "," == self.lookahead(i + 1).type
        ):
            i += 2

        while self.is_selector_token(i) or "," == self.lookahead(i).type:

            if "selector" == self.lookahead(i).type:
                return True

            if "&" == self.lookahead(i + 1).type:
                return True

            if (
                "." == self.lookahead(i).type
                and "ident" == self.lookahead(i + 1).type
            ):
                return True

            if (
                "*" == self.lookahead(i).type
                and "newline" == self.lookahead(i + 1).type
            ):
                return True

            # pseudo-elements
            if (
                ":" == self.lookahead(i).type
                and ":" == self.lookahead(i + 1).type
            ):
                return true

            # #a after an ident and newline
            if (
                "color" == self.lookahead(i).type
                and "newline" == self.lookahead(i - 1).type
            ):
                return true

            if self.looks_like_attribute_selector(i):
                return True

            if (
                "=" == self.lookahead(i).type
                or "function" == self.lookahead(i).type
            ) and "{" == self.lookahead(i + 1).type:
                return False

            # hash values inside properties
            if (
                ":" == self.lookahead(i).type
                and not self.is_pseudo_selector(i + 1)
                and self.line_contains(".")
            ):
                return False

            # the ':' token within braces signifies
            # a selector. ex: "foo{bar:'baz'}"
            if "{" == self.lookahead(i).type:
                brace = True
            elif "}" == self.lookahead(i).type:
                brace = False
            if brace and ":" == self.lookahead(i).type:
                return True

            # '{' preceded by a space is considered a selector.
            # for example "foo{bar}{baz}" may be a property,
            # however "foo{bar} {baz}" is a selector
            if (
                "space" == self.lookahead(i).type
                and "{" == self.lookahead(i + 1).type
            ):
                return True

            # assume pseudo selectors are NOT properties
            # as 'td:th-child(1)' may look like a property
            # and function call to the parser otherwise
            i += 1
            if (
                ":" == self.lookahead(i - 1).type
                and not self.lookahead(i - 1).space
                and self.is_pseudo_selector(i)
            ):
                return True

            # trailing space
            if (
                "space" == self.lookahead(i).type
                and "newline" == self.lookahead(i + 1).type
                and "{" == self.lookahead(i + 2).type
            ):
                return True

            if (
                "," == self.lookahead(i).type
                and "newline" == self.lookahead(i + 1).type
            ):
                return True

        # trailing comma
        if (
            "," == self.lookahead(i).type
            and "newline" == self.lookahead(i + 1).type
        ):
            return True

        # trailing brace
        if (
            "{" == self.lookahead(i).type
            and "newline" == self.lookahead(i + 1).type
        ):
            return True

        # css-style mode, false on ; }
        if self.css > 0 and (
            ";" == self.lookahead(i).type or "}" == self.lookahead(i - 1).type
        ):
            return False

        # trailing separators
        while self.lookahead(i).type not in [
            "indent",
            "outdent",
            "newline",
            "for",
            "if",
            ";",
            "}",
            "eos",
        ]:
            i += 1

        if "indent" == self.lookahead(i).type:
            return True

        return False

    def looks_like_attribute_selector(self, n):
        type = self.lookahead(n).type
        if "=" == type and self.bracketed:
            return True
        return (
            ("ident" == type or "string" == type)
            and "]" == self.lookahead(n + 1).type
            and (
                "newline" == self.lookahead(n + 2).type
                or self.is_selector_token(n + 2)
            )
            and not self.line_contains(":")
            and not self.line_contains("=")
        )

    def looks_like_keyframe(self):
        i = 2
        if self.lookahead(i).type in ["{", "indent", ","]:
            return True
        elif self.lookahead(i).type == "newline":
            i += 1
            while self.lookahead(i).type in ["unit", "newline"]:
                i += 1
            type = self.lookahead(i).type
            return "indent" == type or "{" == type

    def state_allows_selector(self):
        if self.current_state() in [
            "root",
            "atblock",
            "selector",
            "conditional",
            "function",
            "atrule",
            "for",
        ]:
            return True

    def assign_atblock(self, expr):
        try:
            expr.append(self.atblock(expr))
        except Exception:
            self.error(
                "invalid right-hand side operand in assignment, got {peek}"
            )

    def statement(self):
        stmt = self.stmt()
        state = self.prev_state

        # special-case statements since it
        # is not an expression. We could
        # implement postfix conditionals at
        # the expression level, however they
        # would then fail to enclose properties
        if self.allow_postfix:
            self.allow_postfix = False
            state = "expression"

        if state in ["assignment", "expression", "function arguments"]:
            while True:
                op = self.accept(["if", "unless", "for"])
                if op and op.type in ["if", "unless"]:
                    stmt = If(
                        self.expression(),
                        stmt,
                        lineno=self.lineno,
                        column=self.column,
                    )
                    stmt.postfix = true
                    stmt.negate = "unless" == op.type
                    self.accept(";")
                elif op and op.type == "for":
                    val = self.id().name
                    key = None
                    if self.accept(","):
                        key = self.id().name
                    self.expect("in")
                    each = Each(val, key, self.expression())
                    block = Block(
                        self.parent,
                        each,
                        lineno=self.lineno,
                        column=self.column,
                    )
                    block.append(stmt)
                    each.block = block
                    stmt = each
                else:
                    break
        return stmt

    def stmt(self):
        type = self.peek().type
        if type == "keyframes":
            return self.keyframes()
        elif type == "-moz-document":
            return self.mozdocument()
        elif type in [
            "comment",
            "selector",
            "literal",
            "charset",
            "namespace",
            "import",
            "require",
            "extend",
            "media",
            "atrule",
            "ident",
            "scope",
            "supports",
            "unless",
            "function",
            "for",
            "if",
        ]:
            return self.__getattribute__(f"stmt_{type}")()
        elif type in "return":
            return self.return_expression()
        elif type == "{":
            return self.property()
        else:
            if self.state_allows_selector():
                if type in [
                    "color",
                    "~",
                    ">",
                    "<",
                    ":",
                    "&",
                    "&&",
                    "[",
                    ".",
                    "/",
                ]:
                    return self.stmt_selector()
                elif type == "..":
                    # relative reference
                    if "/" == self.lookahead(2).type:
                        return self.stmt_selector()
                elif type == "+":
                    if "function" == self.lookahead(2).type:
                        return self.function_call()
                    else:
                        return self.stmt_selector()
                elif type == "*":
                    return self.property()
                elif type == "unit":
                    if self.looks_like_keyframe():
                        return self.stmt_selector()
                elif type == "-":
                    if "{" == self.lookahead(2).type:
                        return self.property()

        expr = self.expression()
        if expr.is_empty():
            self.error("unexpected {peek}")
        return expr

    def stmt_comment(self):
        node = self.next().value
        self.skip_spaces()
        return node

    def stmt_literal(self):
        return self.expect("literal").value

    def stmt_charset(self):
        self.expect("charset")
        str = self.expect("string").value
        self.allow_postfix = True
        return Charset(str)

    def stmt_namespace(self):
        self.expect("namespace")
        self.skip_spaces_and_comments()
        prefix = self.accept("ident")
        if prefix:
            prefix = prefix.value
        self.skip_spaces_and_comments()

        str = self.accept("string")
        if not str:
            str = self.url()
        self.allow_postfix = True
        return Namespace(str, prefix)

    def stmt_import(self):
        self.expect("import")
        self.allow_postfix = True
        return Import(
            self.expression(), False, lineno=self.lineno, column=self.column
        )

    def stmt_require(self):
        self.expect("require")
        self.allow_postfix = True
        return Import(
            self.expression(), True, lineno=self.lineno, column=self.column
        )

    def stmt_extend(self):
        tok = self.expect("extend")
        selectors = []
        try:
            while True:
                arr = self.selector_parts()
                if arr is None or len(arr) == 0:
                    if self.accept(","):
                        continue
                    else:
                        raise TypeError
                sel = Selector(arr)
                selectors.append(sel)

                if self.peek().type != "!":
                    if self.accept(","):
                        continue
                    else:
                        raise TypeError

                tok = self.lookahead(2)
                if tok.type not in ["ident", "optional"]:
                    continue
                self.skip(["!", "ident"])
                sel.optional = True
                token = self.accept(",")
                if token is None:
                    break
        except TypeError:
            pass

        node = Extend(selectors)
        node.lineno = tok.lineno
        node.column = tok.column
        return node

    def stmt_media(self):
        self.expect("media")
        self.state.append("atrule")
        media = Media(self.queries())
        media.block = self.block(media)
        self.prev_state = self.state[-1]
        self.state.pop()
        return media

    def queries(self):
        queries = QueryList()
        while True:
            self.skip(["comment", "newline", "space"])
            queries.append(self.query())
            self.skip(["comment", "newline", "space"])
            if not self.accept(","):
                break
        return queries

    def query(self):
        query = Query()

        # hash values support
        if self.peek().type == "ident" and (
            self.lookahead(2).type == "." or self.lookahead(2).type == "["
        ):
            self.cond = True
            expr = self.expression()
            self.cond = False
            query.append(Feature(expr.nodes))
            return query

        pred = self.accept(["ident", "not"])
        if pred:
            pred = Literal(pred.value, lineno=self.lineno, column=self.lineno)

            self.skip_spaces_and_comments()
            id = self.accept("ident")
            if id:
                query.type = id.value
                query.predicate = pred
            else:
                query.type = pred
            self.skip_spaces_and_comments()

            if self.accept("&&") is None:
                return query

        while True:
            query.append(self.feature())
            if self.accept("&&") is None:
                break

        return query

    def feature(self):
        self.skip_spaces_and_comments()
        self.expect("(")
        self.skip_spaces_and_comments()
        node = Feature(self.interpolate())
        self.skip_spaces_and_comments()
        self.accept(":")
        self.skip_spaces_and_comments()
        self.in_property = True
        node.expr = self.list()
        self.in_property = False
        self.skip_spaces_and_comments()
        self.expect(")")
        self.skip_spaces_and_comments()
        return node

    def stmt_atrule(self):
        type = self.expect("atrule")
        node = Atrule(type)
        self.skip_spaces_and_comments()
        node.segments = self.selector_parts()
        self.skip_spaces_and_comments()
        tok = self.peek().type
        if tok in ["indent", "{"] or (
            tok == "newline" and self.lookahead(2).type == "{"
        ):
            self.state.append("atrule")
            node.block = self.block(node)
            self.prev_state = self.state[-1]
            self.state.pop()
        return node

    def stmt_scope(self):
        self.expect("scope")
        parts = []
        for literal in self.selector_parts():
            parts.append(str(literal))
        self.selector_scope = "".join(parts).strip()
        return null

    def stmt_supports(self):
        self.expect("supports")
        node = Supports(self.supports_condition())
        self.state.append("atrule")
        node.block = self.block(node)
        self.prev_state = self.state[-1]
        self.state.pop()
        return node

    def supports_condition(self):
        node = self.supports_negation()
        if node is None:
            node = self.supports_op()
        if not node:
            self.cond = True
            node = self.expression()
            self.cond = False
        return node

    def supports_op(self):
        feature = self.supports_feature()
        if feature:
            expr = Expression(lineno=self.lineno, column=self.column)
            expr.append(feature)
            op = self.accept(["&&", "||"])
            while op:
                expr.append(
                    Literal(
                        "and" if op.value == "&&" else "or",
                        lineno=self.lineno,
                        column=self.column,
                    )
                )
                expr.append(self.supports_feature())
                op = self.accept(["&&", "||"])
            return expr
        return None

    def supports_negation(self):
        tok = self.accept("not")
        if tok:
            node = Expression(lineno=self.lineno, column=self.column)
            node.append(Literal("not", lineno=self.lineno, column=self.column))
            node.append(self.supports_feature())
            return node
        return None

    def supports_feature(self):
        self.skip_spaces_and_comments()
        if self.peek().type == "(":
            la = self.lookahead(2).type
            if la in ["ident", "{"]:
                return self.feature()
            else:
                self.expect("(")
                node = Expression(lineno=self.lineno, column=self.column)
                node.append(
                    Literal("(", lineno=self.lineno, column=self.column)
                )
                node.append(self.supports_condition())
                self.expect(")")
                node.append(
                    Literal(")", lineno=self.lineno, column=self.column)
                )
                self.skip_spaces_and_comments()
                return node
        return None

    def stmt_unless(self):
        self.expect("unless")
        self.state.append("conditional")
        self.cond = True
        node = If(
            self.expression(), True, lineno=self.lineno, column=self.column
        )
        self.cond = False
        node.block = self.block(node, False)
        self.prev_state = self.state[-1]
        self.state.pop()
        return node

    def stmt_for(self):
        self.expect("for")
        value = self.id().name
        key = None
        if self.accept(","):
            key = self.id().name
        self.expect("in")
        self.state.append("for")
        self.cond = True
        each = Each(value, key, self.expression())
        self.cond = False
        each.block = self.block(each, False)
        self.prev_state = self.state[-1]
        self.state.pop()
        return each

    def stmt_if(self):
        self.expect("if")
        self.state.append("conditional")
        self.cond = True
        node = If(self.expression(), lineno=self.lineno, column=self.column)
        self.cond = False
        node.block = self.block(node, False)
        self.skip(["newline", "comment"])
        while self.accept("else"):
            if self.accept("if"):
                self.cond = True
                cond = self.expression()
                self.cond = False
                block = self.block(node, False)
                node.elses.append(
                    If(cond, block, lineno=self.lineno, column=self.column)
                )
            else:
                node.elses.append(self.block(node, False))
                break
            self.skip(["newline", "comment"])
        self.prev_state = self.state[-1]
        self.state.pop()
        return node

    def block(self, node, scope=None):
        block = self.parent = Block(
            self.parent, node, lineno=self.lineno, column=self.column
        )

        if scope is False:
            block.scope = False

        self.accept("newline")

        # css style
        if self.accept("{"):
            self.css += 1
            delim = "}"
            self.skip_whitespace()
        else:
            delim = "outdent"
            self.expect("indent")

        while delim != self.peek().type:
            # css-style
            if self.css > 0:
                if self.accept(["newline", "indent"]):
                    continue
                stmt = self.statement()
                self.accept(";")
                self.skip_whitespace()
            else:
                if self.accept("newline"):
                    continue
                # skip useless indents and comments
                next = self.lookahead(2).type
                if "indent" == self.peek().type and next in [
                    "outdent",
                    "newline",
                    "comment",
                ]:
                    self.skip(["indent", "outdent"])
                    continue
                if "eos" == self.peek().type:
                    return block
                stmt = self.statement()
                self.accept(";")
            if not stmt:
                self.error("unexpected token {peek} in block")
            block.append(stmt)

        # css-style
        if self.css > 0:
            self.skip_whitespace()
            self.expect("}")
            self.skip_spaces()
            self.css -= 1
        else:
            self.expect("outdent")

        self.parent = block.parent

        return block

    def stmt_selector(self):
        scope = self.selector_scope
        group = Group(lineno=self.lineno, column=self.column)
        is_root = self.current_state() == "root"

        while True:
            self.accept("newline")  # clobber newline after ','
            arr = self.selector_parts()

            # push the selector
            if is_root and scope:
                arr.appendleft(
                    Literal(
                        f"{scope} ", lineno=self.lineno, column=self.column
                    )
                )
            if len(arr) > 0:
                selector = Selector(arr)
                selector.lineno = arr[0].lineno
                selector.column = arr[0].column
                group.append(selector)

            if self.accept([",", "newline"]) is None:
                break

        if "selector-parts" == self.current_state():
            return group.nodes

        self.state.append("selector")
        group.block = self.block(group)
        self.prev_state = self.state[-1]
        self.state.pop()

        return group

    def selector_parts(self) -> deque:
        """Selector candidates, stitched together to form a selector."""
        arr = deque()
        while True:
            tok = self.selector_token()
            if tok:
                if tok.type == "{":
                    self.skip_spaces()
                    expr = self.expression()
                    self.skip_spaces()
                    self.expect("}")
                    arr.append(expr)
                elif tok.type == self.prefix and ".":
                    literal = Literal(
                        tok.value + self.prefix,
                        lineno=self.lineno,
                        column=self.column,
                    )
                    literal.prefixed = True
                    arr.append(literal)
                elif tok.type == "comment":
                    # ignore comments
                    pass
                elif tok.type in ["color", "unit"]:
                    arr.append(
                        Literal(
                            tok.value.raw,
                            lineno=self.lineno,
                            column=self.column,
                        )
                    )
                elif tok.type == "space":
                    arr.append(
                        Literal(" ", lineno=self.lineno, column=self.column)
                    )
                elif tok.type == "function":
                    arr.append(
                        Literal(
                            f"{tok.value.name}(",
                            lineno=self.lineno,
                            column=self.column,
                        )
                    )
                elif tok.type == "ident":
                    if hasattr(tok.value, "name") and tok.value.name:
                        arr.append(
                            Literal(
                                f"{tok.value.name}",
                                lineno=self.lineno,
                                column=self.column,
                            )
                        )
                    else:
                        arr.append(
                            Literal(
                                f"{tok.value.string}",
                                lineno=self.lineno,
                                column=self.column,
                            )
                        )
                else:
                    arr.append(
                        Literal(
                            tok.value, lineno=self.lineno, column=self.column
                        )
                    )
                    if tok.space:
                        arr.append(Literal(" "))
            else:
                break
        return arr

    def assignment(self):
        name = self.id().name
        op = self.accept(["=", "?=", "+=", "-=", "*=", "/=", "%="])
        if op:
            self.state.append("assignment")
            expr = self.list()
            # @block support
            if expr.is_empty():
                self.assign_atblock(expr)
            node = Ident(name, expr)
            self.prev_state = self.state[-1]
            self.state.pop()

            if op.type == "?=":
                defined = BinOp("is defined", node)
                lookup = Expression(lineno=self.lineno, column=self.column)
                lookup.append(Ident(name))
                node = Ternary(defined, lookup, node)
            elif op.type in ["+=", "-=", "*=", "/=", "%="]:
                node.value = BinOp(op.type[0], Ident(name), expr)

        return node

    def stmt_ident(self):  # noqa: C901
        i = 2
        la = self.lookahead(i).type

        while "space" == la:
            i += 1
            la = self.lookahead(i).type

        go_on = False

        if la in ["=", "?=", "-=", "+=", "*=", "/=", "%="]:
            # assignment
            return self.assignment()

        if la == ".":
            # member
            if "space" == self.lookahead(i - 1).type:
                return self.stmt_selector()
            if self._ident == self.peek():
                return self.id()
            i += 1
            some_tokens = ["[", ",", "newline", "indent", "eos"]
            while (
                "=" != self.lookahead(i).type
                and self.lookahead(i).type not in some_tokens
            ):
                i += 1
            if "=" == self.lookahead(i).type:
                self._ident = self.peek()
                return self.expression()
            elif self.looks_like_selector() and self.state_allows_selector():
                return self.stmt_selector()
            else:
                go_on = True

        if la == "[" or go_on:
            # assignment []=
            if self._ident == self.peek():
                return self.id()
            while (
                "]" != self.lookahead(i).type
                and "selector" != self.lookahead(i + 1).type
                and "eos" != self.lookahead(i + 1).type
            ):
                i += 1
            i += 1
            if "=" == self.lookahead(i).type:
                self._ident = self.peek()
                return self.expression()
            elif self.looks_like_selector() and self.state_allows_selector():
                return self.stmt_selector()
            go_on = True

        # operation
        if (
            la
            in [
                "-",
                "+",
                "/",
                "*",
                "%",
                "**",
                "&&",
                "||",
                ">",
                "<",
                ">=",
                "<=",
                "!=",
                "==",
                "?",
                "in",
                "is a",
                "is defined",
            ]
            or go_on
        ):
            if self._ident == self.peek():
                return self.id()
            else:
                self._ident = self.peek()
                if self.current_state() in ["for", "selector"]:
                    return self.property()
                if self.current_state() in ["root", "atblock", "atrule"]:
                    if la == "[":
                        return self.subscript()
                    else:
                        return self.stmt_selector()
                if self.current_state() in ["function", "conditional"]:
                    if self.looks_like_selector():
                        return self.stmt_selector()
                    else:
                        return self.expression()
                if self.operand:
                    return self.id()
                else:
                    return self.expression()

        if self.current_state() == "root":
            return self.stmt_selector()
        elif self.current_state() in [
            "for",
            "selector",
            "function",
            "conditional",
            "atblock",
            "atrule",
        ]:
            return self.property()
        else:
            id = self.id()
            if "interpolation" == self.previous_state():
                id.mixin = True
            return id

    def property(self):
        if self.looks_like_selector(True):
            return self.stmt_selector()

        # property
        ident = self.interpolate()
        prop = Property(ident, lineno=self.lineno, column=self.column)
        ret = prop

        # optional ':'
        self.accept("space")
        if self.accept(":"):
            self.accept("space")

        self.state.append("property")
        self.in_property = True
        prop.expr = self.list()
        if len(prop.expr) == 0:
            ret = ident[0]
        self.in_property = False
        self.allow_postfix = True
        self.prev_state = self.state[-1]
        self.state.pop()

        # optional ';'
        self.accept(";")

        return ret

    def interpolate(self):
        node = None
        segs = []

        star = self.accept("*")
        if star:
            segs.append(Literal("*", lineno=self.lineno, column=self.column))

        while True:
            if self.accept("{"):
                self.state.append("interpolation")
                segs.append(self.expression())
                self.expect("}")
                self.prev_state = self.state[-1]
                self.state.pop()
            elif self.accept("-"):
                segs.append(
                    Literal("-", lineno=self.lineno, column=self.column)
                )
            else:
                node = self.accept("ident")
                if node:
                    segs.append(node.value)
                else:
                    break

        # empty segment list
        if len(segs) == 0:
            self.expect("ident")

        return segs

    # fixme!
    def expression(self):
        """negation+"""
        expr = Expression(lineno=self.lineno, column=self.column)
        self.state.append("expression")
        while True:
            node = self.negation()
            if node is None:
                self.error("unexpected token {peek} in expression")
            if node.node_name == "nothing":
                break
            expr.append(node)
        self.prev_state = self.state[-1]
        self.state.pop()
        if len(expr.nodes) > 0:
            expr.lineno = expr.nodes[0].lineno
            expr.column = expr.nodes[0].column
        return expr

    def negation(self):
        if self.accept("not"):
            return UnaryOp("!", self.negation())
        return self.ternary()

    def ternary(self):
        node = self.logical()
        if self.accept("?"):
            true_expr = self.expression()
            self.expect(":")
            false_expr = self.expression()
            node = Ternary(node, true_expr, false_expr)
        return node

    def list(self):
        """expression (',' expression)*"""
        node = self.expression()

        while self.accept(","):
            if node.is_list:
                # todo: fixme was list, now node
                node.append(self.expression())
            else:
                list = Expression(true, lineno=self.lineno, column=self.column)
                list.append(node)
                list.append(self.expression())
                node = list

        return node

    def logical(self):
        node = self.typecheck()
        while True:
            op = self.accept(["&&", "||"])
            if op:
                node = BinOp(op.type, node, self.typecheck())
            else:
                break
        return node

    def typecheck(self):
        node = self.equality()
        while True:
            op = self.accept("is a")
            if op:
                self.operand = True
                if not node:
                    self.error(
                        f'illegal unary "{op}", ' f"missing left-hand operand"
                    )
                node = BinOp(op.type, node, self.equality())
                self.operand = False
            else:
                break
        return node

    def equality(self):
        node = self.inn()
        while True:
            op = self.accept(["==", "!="])
            if op:
                self.operand = True
                if node is None:
                    self.error(
                        f'illegal unary "{op}", ' f"missing left-hand operand"
                    )
                node = BinOp(op.type, node, self.inn())
                self.operand = False
            else:
                break
        return node

    def inn(self):
        node = self.relational()
        while self.accept("in"):
            self.operand = True
            if not node:
                self.error('illegal unary "in", ' "missing left-hand operand")
            node = BinOp("in", node, self.relational())
            self.operand = False
        return node

    def relational(self):
        node = self.range()
        op = self.accept([">=", "<=", "<", ">"])
        if op:
            self.operand = True
            if not node:
                self.error(
                    f'illegal unary "{op}", ' f"missing left-hand operand"
                )
            node = BinOp(op.type, node, self.range())
            self.operand = False
        return node

    def stmt_function(self):
        parens = 1
        i = 2

        # Lookahead and determine if we are dealing
        # with a function call or definition. Here
        # we pair parens to prevent false negatives

        # todo: clean this!
        try:
            tok = self.lookahead(i)
            i += 1
            while tok:
                if tok.type in ["function", "("]:
                    parens += 1
                    tok = self.lookahead(i)
                    i += 1
                    continue
                elif tok.type == ")":
                    parens -= 1
                    if parens == 0:
                        raise TypeError  # which is good
                    tok = self.lookahead(i)
                    i += 1
                    continue
                elif tok.type == "oes":
                    self.error('failed to find closing paren ")"')
                tok = self.lookahead(i)
                i += 1
        except TypeError:
            pass

        if self.current_state() == "expression":
            return self.function_call()
        else:
            if self.looks_like_function_definition(i):
                return self.function_definition()
            else:
                return self.expression()

    def range(self):
        node = self.additive()
        op = self.accept(["...", ".."])
        if op:
            self.operand = True
            if not node:
                self.error(f'illegal unary "{op}", missing left-hand operand')
            node = BinOp(op.value, node, self.additive())
            self.operand = False
        return node

    def additive(self):
        node = self.multiplicative()
        op = self.accept(["+", "-"])
        while op:
            self.operand = True
            node = BinOp(op.type, node, self.multiplicative())
            self.operand = False
            op = self.accept(["+", "-"])
        return node

    def multiplicative(self):
        node = self.defined()
        op = self.accept(["**", "*", "/", "%"])
        while op:
            self.operand = True
            if "/" == op.value and self.in_property and self.parens == 0:
                self.stash.append(
                    Token(
                        "literal",
                        Literal("/", lineno=self.lineno, column=self.column),
                    )
                )
                self.operand = False
                return node
            else:
                if not node:
                    self.error(
                        f'illegal unary "{op}", ' f"missing left-hand operand"
                    )
                node = BinOp(op.type, node, self.defined())
                self.operand = False
            op = self.accept(["**", "*", "/", "%"])
        return node

    def defined(self):
        node = self.unary()
        if self.accept("is defined"):
            if not node:
                self.error(
                    f'illegal unary "is defined", '
                    f"missing left-hand operand"
                )
            node = BinOp("is defined", node)
        return node

    def unary(self):
        op = self.accept(["!", "~", "+", "-"])
        if op:
            self.operand = True
            node = self.unary()
            if node is None:
                self.error(f'illegal unary "{op}"')
            node = UnaryOp(op.type, node)
            self.operand = False
            return node
        return self.subscript()

    def subscript(self):
        node = self.member()
        while self.accept("["):
            node = BinOp("[]", node, self.expression())
            self.expect("]")
        if self.accept("="):
            node.op += "="
            node.value = self.list()
            # @block suppprt
            if node.value.is_empty():
                self.assign_atblock(node.value)
        return node

    def member(self):
        node = self.primary()
        if node and node.node_name != "nothing":
            while self.accept("."):
                id = Ident(self.expect("ident").value.string)
                node = Member(node, id)
            self.skip_spaces()
            if self.accept("="):
                node.value = self.list()
                # @block support
                if node.value.is_empty():
                    self.assign_atblock(node.value)
        return node

    def primary(self):
        self.skip_spaces()

        # parenthesis
        if self.accept("("):
            self.parens += 1
            expr = self.expression()
            paren = self.expect(")")
            self.parens -= 1
            if self.accept("%"):
                expr.append(Ident("%"))
            tok = self.peek()
            # (1 + 2)px, (1 + 2)em, etc.
            if (
                not paren.space
                and "ident" == tok.type
                and tok.value.string in units
            ):
                expr.append(Ident(tok.value.string))
                self.next()
            return expr

        tok = self.peek()

        # primitive
        if tok.type in [
            "null",
            "unit",
            "color",
            "string",
            "literal",
            "boolean",
            "comment",
        ]:
            return self.next().value
        elif not self.cond and tok.type == "{":
            return self.object()
        elif tok.type == "atblock":
            return self.atblock()
        # property lookup
        elif tok.type == "atrule":
            id = Ident(self.next().value)
            id.property = True
            return id
        elif tok.type == "ident":
            return self.stmt_ident()
        elif tok.type == "function":
            if tok.anonymous:
                return self.function_definition()
            else:
                return self.function_call()

        return Nothing()

    def function_definition(self):
        name = self.expect("function").value.name

        # params
        self.state.append("function params")
        self.skip_whitespace()
        params = self.params()
        self.skip_whitespace()
        self.expect(")")
        self.prev_state = self.state[-1]
        self.state.pop()

        # body
        self.state.append("function")
        fn = Function(
            name, params, builtin=False, lineno=self.lineno, column=self.column
        )
        fn.block = self.block(fn)
        self.prev_state = self.state[-1]
        self.state.pop()
        return Ident(name, fn)

    def id(self):
        tok = self.expect("ident")
        self.accept("space")
        return tok.value

    def atblock(self, node=None):
        if node is None:
            self.expect("atblock")
        n = Atblock()
        self.state.append("atblock")
        n.block = self.block(n, scope=False)
        self.prev_state = self.state[-1]
        self.state.pop()
        return n

    def return_expression(self):
        self.expect("return")
        expr = self.expression()
        if expr.is_empty():
            return ReturnNode()
        else:
            return ReturnNode(expr)
        pass

    def keyframes(self):
        tok = self.expect("keyframes")
        self.skip_spaces_and_comments()
        keyframes = Keyframes(self.selector_parts(), tok.value)
        self.skip_spaces_and_comments()

        # block
        self.state.append("atrule")
        keyframes.block = self.block(keyframes)
        self.prev_state = self.state[-1]
        self.state.pop()

        return keyframes

    def args(self):
        args = Arguments()
        while True:
            if self.peek().type == "ident" and self.lookahead(2).type == ":":
                keyword = self.next().value.string
                self.expect(":")
                args.map[keyword] = self.expression()
            else:
                args.append(self.expression())
            if self.accept(",") is None:
                break
        return args

    def url(self):
        self.expect("function")
        self.state.append("function")
        args = self.args()
        self.expect(")")
        self.prev_state = self.state[-1]
        self.state.pop()
        return Call("url", args, lineno=self.lineno, column=self.column)

    def mozdocument(self):
        self.expect("-moz-document")
        mozdocument = Atrule("-moz-document")
        calls = []
        while True:
            self.skip_spaces_and_comments()
            calls.append(self.function_call())
            self.skip_spaces_and_comments()
            if self.accept(",") is None:
                break
        joined_calls = [str(call) for call in calls]
        mozdocument.segments = [
            Literal(
                ", ".join(joined_calls), lineno=self.lineno, column=self.column
            )
        ]
        self.state.append("atrule")
        mozdocument.block = self.block(mozdocument, False)
        self.prev_state = self.state[-1]
        self.state.pop()
        return mozdocument

    def function_call(self):
        with_block = self.accept("+")
        if self.peek().value.name == "url":
            return self.url()
        name = self.expect("function").value.name
        self.state.append("function arguments")
        self.parens += 1
        args = self.args()
        self.expect(")")
        self.parens -= 1
        self.prev_state = self.state[-1]
        self.state.pop()
        call = Call(name, args, lineno=self.lineno, column=self.column)
        if with_block:
            self.state.append("function")
            call.block = self.block(call)
            self.prev_state = self.state[-1]
            self.state.pop()
        return call

    def params(self):
        params = Params()
        tok = self.accept("ident")
        while tok:
            self.accept("space")
            node = tok.value
            params.append(node)
            if self.accept("..."):
                node.rest = True
            elif self.accept("="):
                node.value = self.expression()
            self.skip_whitespace()
            self.accept(",")
            self.skip_whitespace()
            tok = self.accept("ident")
        params.lineno = self.lineno
        params.column = self.column
        return params

    def object(self):
        comma = None
        obj = ObjectNode({})
        self.expect("{")
        self.skip_whitespace()

        while True:
            token = self.next()
            if token and token.type == "}":
                break
            if token.type in ["comment", "newline"]:
                continue
            if not comma and token.type == ",":
                continue
            if token.type in ["ident", "string"]:
                id = token
            if id is None:
                self.error('expected "ident" or "string", got {peek}')
            id = id.value.hash()
            self.skip_spaces_and_comments()
            self.expect(":")
            value = self.expression()
            obj.set(id, value)
            comma = self.accept(",")
            self.skip_whitespace()

        return obj

    def _operation(self, la):
        # operation
        # prevent cyclic .ident, return literal
        if self._ident == self.peek():
            return self.id()
        else:
            self._ident = self.peek()
            if self.current_state() in ["for", "selector"]:
                return self.property()
            elif self.current_state() in ["root", "atblock", "atrule"]:
                if "[" == la:
                    return self.subscript()
                else:
                    return self.stmt_selector()
            elif self.current_state() in ["function", "conditional"]:
                if self.looks_like_selector():
                    return self.stmt_selector()
                else:
                    return self.expression()
            else:
                # do not disrupt the ident when an operand
                if self.operand:
                    return self.id()
                else:
                    return self.expression()



import math as m

from stilus.nodes.string import String
from stilus.nodes.unit import Unit
from stilus.utils import assert_type, assert_string, stilus_round


def math(n: Unit, fn: String, evaluator=None):
    """Apply Math ``fn`` to ``n``"""
    assert_type(n, "unit", "n")
    assert_string(fn, "fn")

    if fn.string == "round":
        return Unit(int(stilus_round(n.value)), n.type)

    return Unit(m.__getattribute__(fn.string)(n.value), n.type)



from stilus.nodes.color import RGBA
from stilus.utils import assert_color


def blend(top, bottom=None, evaluator=None):
    assert_color(top)
    top = top.rgba()
    if not bottom:
        bottom = RGBA(255, 255, 255, 1)
    assert_color(bottom)
    bottom = bottom.rgba()

    return RGBA(
        top.r * top.a + bottom.r * (1 - top.a),
        top.g * top.a + bottom.g * (1 - top.a),
        top.b * top.a + bottom.b * (1 - top.a),
        top.a + bottom.a - top.a * bottom.a,
    )
